#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <limits.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <glob.h>
#include <unistd.h>
#include <sysfs/libsysfs.h>
#include <sys/wait.h>
#include <ctype.h>

#include "settings.h"
#include "lib.h"

extern const char *__progname;
struct envglobal g;

void program_version( FILE *output )
{
    fprintf(output, PACKAGE" kernel module switcher v%s by Sergey Turchin <zerg@altlinux.org>\n", PACKAGE_VERSION);
}

void program_usage( FILE *output )
{
    program_version( output );
    fprintf(output,
"Usage:\n"
"  nvidia -h				-- display command usage\n"
"  nvidia -V				-- display program version\n"
"  nvidia -v, --verbose			-- enable verbose output\n"
);
}

int main(int argc, char **argv)
{
    struct xinflist *paired_xinfs = 0;
    struct nvdevlist *nvidia_cards = 0;
    static struct option long_options[] = { 
	{ "verbose", 0, NULL, 'v' }, 
	{ "help", 0, NULL, 'h' }, 
	{ "version", 0, NULL, 'V'},
	{ NULL, 0, NULL, 0 } };

    init_global();

    while(1)
    {
	int c = getopt_long( argc, argv, "hvV", long_options, NULL );
	if( c == -1 ) 
    	    break;
	switch( c )
	{
	    case 'v':
		g.o_verbose = 1;
		break;
	    case 'V':
		program_version( stdout );
		exit(0);
	    case 'h':
		program_usage( stdout );
		exit(0);
	    default: 
		program_usage( stderr );
		exit(1);
	}
    }

    get_kernel_ver();
    get_last_kmod_ver();

    { /* find paired xinfs */
	glob_t globbuf;
	size_t i;

	if( glob(XINF_DIR"nvidia-*.xinf", GLOB_NOESCAPE, NULL, &globbuf) == 0 )
	{	
	    for(i=0; i < globbuf.gl_pathc; i++)
	    {
		const char *path = *(globbuf.gl_pathv);
		char *start;
		start = strstr(path, "-");
		if( start )
		{
			char *end;
			start++;
			end = strstr(start, ".xinf");
			if( end )
			{
			    int len = strlen(start)-strlen(end);
			    if( len < VER_LEN_MAX && len > 0 );
			    {
				char nv_kmod_file[PATH_MAX];
				char nv_ver_str[VER_LEN_MAX];
				strncpy(nv_ver_str, start, len);
				nv_ver_str[len] = '\0';
				strcpy(nv_kmod_file, NV_KMOD_DIR);
				strcat(nv_kmod_file, g.kernel_ver_str);
				strcat(nv_kmod_file, "-");
				strcat(nv_kmod_file, nv_ver_str);
				if( access(nv_kmod_file, F_OK) == 0 )
				{
				    struct xinflist *xinf = malloc(sizeof(struct xinflist));
				    strcpy(xinf->xinf, path);
				    strcpy(xinf->module, nv_kmod_file);
				    strcpy(xinf->version, nv_ver_str);
				    xinf->supported = 0;
				    xinf->next = paired_xinfs;
				    paired_xinfs = xinf;
				    if( g.o_verbose )
					fprintf(stderr, "%s: found %s\n", __progname, path);
				}
			    }
			}
		}
		globbuf.gl_pathv++;
	    }
	}
    }

    { /* scan pci */
	int other_videocards = 0;
	struct sysfs_bus *bus;
	bus = sysfs_open_bus("pci");
	if( bus != NULL )
	{
	    struct dlist *devlist;
	    struct sysfs_device *curdev;
	    devlist = sysfs_get_bus_devices(bus);
	    if( devlist )
	    {
		dlist_for_each_data(devlist, curdev, struct sysfs_device)
		{
		    struct sysfs_attribute *class_attr;
		    class_attr = sysfs_get_device_attr(curdev, "class");
		    if( class_attr != NULL )
		    {
			long int class_id;
			class_id = strtol(class_attr->value,NULL,16);
			if( class_id == 0x030000 )
			{
			    struct sysfs_attribute *vendor_attr;
			    struct sysfs_attribute *device_attr;
			    vendor_attr = sysfs_get_device_attr(curdev, "vendor");
			    device_attr = sysfs_get_device_attr(curdev, "device");
			    if( vendor_attr != NULL && device_attr != NULL )
			    {
				unsigned int vendor_id, device_id;
				vendor_id = strtol(vendor_attr->value,NULL,16);
				device_id = strtol(device_attr->value,NULL,16);
				if( vendor_id == 0x10de || vendor_id == 0x12d2 )
				{
				    struct nvdevlist *nvdev = malloc(sizeof(struct nvdevlist));
				    nvdev->vendor = vendor_id;
				    nvdev->device = device_id;
				    nvdev->next = nvidia_cards;
				    nvidia_cards = nvdev;
				    if( g.o_verbose )
					fprintf(stderr, "%s: found device %08x:%08x\n", __progname, vendor_id, device_id);
				}
				else
				{
				    if( g.o_verbose )
					fprintf(stderr, "%s: found other device %08x:%08x\n", __progname, vendor_id, device_id);
				}
			    }
			    else
			    {
				other_videocards++;
			    }
			}
		    }
		}
	    }
	}
	else
	    fprintf(stderr,"%s: failed to scan pci via sysfs\n", __progname);
	/* remember non-nvidia cards quantity */
	g.other_videocards = other_videocards;
    }

    /* grep xinfs */
    if( paired_xinfs && nvidia_cards )
    {
	struct xinflist *xinf;
	for(xinf = paired_xinfs; xinf; xinf = xinf->next)
	{
	    struct nvdevlist *nvdev;
	    for( nvdev = nvidia_cards; nvdev; nvdev=nvdev->next )
	    {
		FILE *xinffile;
		char pattern[64];
		sprintf(pattern, "pcivideo:v%08xd%08x", nvdev->vendor, nvdev->device);
		xinffile = fopen(xinf->xinf, "r");
		if( xinffile )
		{
		    while( !feof(xinffile) )
		    {
			char buf[LINE_MAX];
			char *line;
			int len;
			line = fgets(buf, LINE_MAX-1, xinffile);
			if( line )
			{
			    len = strlen(line);
			    if( buf[len-1] == '\n' )
			    {
				len--;
				buf[len] = '\0';
			    }
			    if( strcasestr(line, pattern) )
				xinf->supported++;
			}
		    }
		    fclose(xinffile);
		}
	    }
	}
    }

    /* select best version */    
    if( paired_xinfs && nvidia_cards )
    {
	int max_support = 0;
	struct xinflist *xinf;
	for(xinf = paired_xinfs; xinf; xinf = xinf->next)
	{
	    if( xinf->supported > 0
		&& xinf->supported >= max_support
		&& version_compare(string_to_version(xinf->version), string_to_version(g.new_kmod_ver_str)) > 0 )
	    {
		max_support = xinf->supported;
		strcpy(g.new_kmod_ver_str, xinf->version);
	    }
	}
	if( strlen(g.new_kmod_ver_str) > 0 )
	{
	    if( g.o_verbose )
		printf("%s: best kernel module version - %s\n", __progname, g.new_kmod_ver_str);
	}
    }

    /* unload kernel module */
    if( strlen(g.new_kmod_ver_str) > 0 )
    {
	int sysret = 0;
	{ /* read loaded module version */
	    FILE *proc_modules;
	    proc_modules = fopen("/proc/modules", "r");
	    if( proc_modules )
	    {
		while( !feof(proc_modules) )
		{
		    char buf[256];
		    char *line;
		    line = fgets(buf, 255, proc_modules);
		    if( line && strlen(line) > 6 && strncmp("nvidia", line, 6) == 0 && isspace(buf[6]) )
		    {
			read_version_from_proc(g.proc_kmod_ver_str, "/proc/driver/nvidia/version");
		    }
		}
		fclose(proc_modules);
	    }
	}
	if( strlen(g.new_kmod_ver_str) > 0
	    && strlen(g.proc_kmod_ver_str) > 0
	    && strcmp(g.new_kmod_ver_str, g.proc_kmod_ver_str) != 0 )
	{
	    sysret = system("/sbin/modprobe -r nvidia");
	    if( sysret != 0 )
	    {
	        fprintf(stderr, "%s: Warning! You must reboot computer or shutdown X-server and unload kernel module by command `/sbin/modprobe -r nvidia`\n", __progname);
	    }
	    else
	    {
	        if( g.o_verbose )
		    fprintf(stderr, "%s: unload kernel module\n", __progname);
	    }
	}
	else
	{
	    if( g.o_verbose )
	        fprintf(stderr, "%s: don't unload kernel module\n", __progname);
	}
    }

    /* switch kernel module */
    if( strlen(g.new_kmod_ver_str) > 0 )
    {
	char kmod_dir[PATH_MAX];
	char nv_kmod_file[PATH_MAX];
	char version_file_path[PATH_MAX];
	int depmod_sysret = 0;
	int is_kmod_change = 0;

	is_kmod_change = strcmp(g.last_kmod_ver_str, g.new_kmod_ver_str) != 0;

	strcpy(kmod_dir, K_MOD_PREFIX);
	strcat(kmod_dir, g.kernel_ver_str);
	strcat(kmod_dir, "/nVidia/");

	strcpy(nv_kmod_file, g.kernel_ver_str);
	strcat(nv_kmod_file, "-");
	strcat(nv_kmod_file, g.new_kmod_ver_str);

	if( is_kmod_change )
	    printf("%s: switching kernel module form %s to %s\n", __progname, g.last_kmod_ver_str, g.new_kmod_ver_str);
	mk_symlink_from_to(0, nv_kmod_file, NV_KMOD_DIR, "nvidia.ko", kmod_dir);

	strcpy(version_file_path, K_MOD_PREFIX);
	strcat(version_file_path, g.kernel_ver_str);
	strcat(version_file_path, K_MOD_VER_SUFFIX);
	write_version_to_file(g.new_kmod_ver_str, version_file_path);

	if( is_kmod_change )
	    depmod_sysret = system("/sbin/depmod -a");
	if( depmod_sysret != 0 )
	{
	    /* something */
	}
    }

    /* create xorg.conf */
    if( g.other_videocards <= 0 && strlen(g.new_kmod_ver_str) > 0 )
    {
	int nouveau_loaded = 0;
	{ /* check for nouveau kernel module */
	    FILE *proc_modules;
	    proc_modules = fopen("/proc/modules", "r");
	    if( proc_modules )
	    {
		while( !feof(proc_modules) )
		{
		    char buf[256];
		    char *line;
		    line = fgets(buf, 255, proc_modules);
		    if( line && strlen(line) > 7 && strncmp("nouveau", line, 7) == 0 && isspace(buf[7]) )
		    {
			nouveau_loaded = 1;
		    }
		}
		fclose(proc_modules);
	    }
	}
	if( !nouveau_loaded && access("/etc/X11/xorg.conf", F_OK) != 0 )
	{
	    int sysret = 0;
	    sysret = system("xsetup-monitor --create-only --driver nvidia");
	    if( sysret != 0 )
		sysret = system("nvidia-xconfig");
	}
    }

    restore_symlinks();

/* used with old scheme
    if( mk_symlink(0, "nvidia.xinf", NV_ETCLIB_SYML_DIR) == 0 )
	mk_symlink_from_to(0, "null", "/dev/", "nvidia.xinf", NV_ETCLIB_SYML_DIR);
*/
    return(0);
}
